from ..types.methods import Method
from typing import Callable, List
from vbml import Pattern

import re


class Handler:

    def __init__(self, handler: Callable, event_type: Method):
        self.event_type = event_type
        self.handler = handler

    async def notify_handler(self, event):
        await self.handler(event)


class SelfHandler(Handler):

    def __init__(
            self,
            handler: Callable,
            event_type: Method,
            commands: List[str],
            lower: bool
    ):
        self.event_type: Method = event_type
        self.handler: Callable = handler
        self.commands: List[str] = commands
        self.lower: bool = lower
        self.patterns: List[Pattern] = []
        for i in self.commands:
            self.patterns.append(
                Pattern(
                    text=i,
                    flags=re.IGNORECASE if lower else None
                )
            )
        super().__init__(handler, event_type)


from aiohttp import web
from ..utils import logger
from ..utils import errors


class IrisHandler(web.View):

    _methods = ["ping", "subscribeSignals", "banExpired",
                "addUser", "deleteMessages", "hireApi",
                "deleteMessagesFromUser", "printBookmark",
                "forbiddenLinks", "sendSignal", "toGroup",
                "sendMySignal", "banGetReason", "bindChat",
                "ignoreMessages"]

    async def get(self):
        """ Метод принимающий GET запросы.
        Но он нам нахер не нужен, поэтому кидаем
        HTTPForbidden адресату.
        :return: -> None
        """
        raise web.HTTPForbidden

    async def post(self):
        event = await self.request.json()
        method = event.get("method")
        if not method:
            return web.json_response(data=errors[1])

        logger.debug(f"{event}")

        secret = event.get("secret")
        user_id = event.get("user_id")

        if secret != self.request.app["secret"]:
            logger.warning(f"Got invalid secret string: {secret}")
            return web.json_response(data=errors[3])

        if user_id != self.request.app["user_id"]:
            logger.warning(f"Got invalid user_id: {user_id}")
            return web.json_response(data=errors[3])

        await self.request.app["dp"].process_events([event])
        if method in self._methods:
            return web.Response(text="ok")
        logger.warning(f"Detected new unknown method: {method}")
        return web.json_response(data=errors[2])


def get_app(dp, secret: str, user_id: int):
    app = web.Application(logger=logger, loop=dp.loop)
    app["dp"] = dp
    app["secret"] = secret
    app["user_id"] = user_id
    return app


def run_app(app: web.Application, host: str, port: int, path: str):
    app.router.add_view(path=path, handler=IrisHandler)
    web.run_app(app, host=host, port=port)


from ..types.methods import Method
from ..types.events import Event
from ..types import objects
from ..utils import logger
from ..utils import sub_string

from . import server
from ._status import LoggerLevel
from vbml import Patcher

import asyncio
import typing
import sys


class Dispatcher:

    def __init__(
            self,
            secret: str,
            user_id: int,
            debug: typing.Union[str, bool] = True,
            log_to_path: typing.Union[str, bool] = None,
            *,
            loop: asyncio.AbstractEventLoop = None
    ):
        self.loop = loop if loop else asyncio.get_event_loop()
        self.secret: str = secret
        self.debug: bool = debug
        self.user_id: int = user_id
        self.event: Event = Event()
        self.patcher: Patcher = Patcher()

        if isinstance(debug, bool):
            debug = "INFO" if debug else "ERROR"

        self.logger = LoggerLevel(debug)

        logger.remove()
        logger.add(
            sys.stderr,
            colorize=True,
            format="<level>[<blue>IrisPY</blue>] {message}</level> <white>[TIME {time:HH:MM:ss}]</white>",
            filter=self.logger,
            level=0
        )
        logger.level("INFO", color="<white>")
        logger.level("ERROR", color="<red>")
        if log_to_path:
            logger.add(
                "log_{time}.log" if log_to_path is True else log_to_path,
                rotation="100 MB",
            )

        logger.debug("Initialized dispatcher with SECRET: <{}> USER_ID: <{}>".format(
            secret, user_id
        ))

    async def process_event(self, event: dict):
        """ Функция, отвечающая за обработку эвента.
        :param event: -> dict
        :return: -> None
        """
        _event = await self.get_event_type(event)
        for handler in self.event.handlers:
            if handler.event_type.value == _event.method:
                try:
                    await handler.notify_handler(_event)
                    logger.info("-> NEW EVENT {} FROM CHAT {}".format(
                        _event.method, _event.object.chat
                    ))
                except Exception as e:
                    logger.exception(f"Error in handler: {e}")

    async def process_self_event(self, event: dict):
        """ Обработка эвентов: "sendSignal", "SendMySignal".
        :return:
        """
        _event = await self.get_event_type(event)
        for handler in self.event.event_handlers:
            if handler.event_type.value == _event.method:
                try:
                    await self.validation(
                        func=handler.handler,
                        event=_event,
                        text=sub_string(_event.message.text),
                        patterns=handler.patterns
                    )
                    logger.info(
                        "-> NEW EVENT {} FROM CHAT {}".format(
                            _event.method, _event.object.chat
                        ))
                except Exception as e:
                    logger.exception(f"Error in handler: {e}")

    async def process_events(self, events: typing.List[dict]):
        for event in events:
            if event["method"] not in ("sendMySignal", "sendSignal"):
                self.loop.create_task(self.process_event(event))
            else:
                self.loop.create_task(self.process_self_event(event))

    async def validation(
            self,
            func: typing.Callable,
            event: Method,
            text: str,
            patterns
    ):
        kwargs = {}
        for i in patterns:
            result = self.patcher.check(text, i)
            if result == {}:
                return await func(event)
            if result:
                kwargs.update(**result)
        logger.debug(f"Validation's result: {kwargs}")
        if kwargs != {}:
            return await func(event, **kwargs)

    def run_app(self, host: str = "0.0.0.0", port: int = 8080, path: str = "/"):
        """
        :param host: IP адресс, где будет запущен сервер: Пример: "127.0.0.1"
        :param port: Порт, на котором будет запущен сервер: Пример 8000
        :param path: Путь, куда Ирис будет отсылать POST запросы: Пример "/bot"
        :return: -> None
        """
        app = server.get_app(self, self.secret, self.user_id)
        logger.info("Handling successfully started. Press Ctrl+C to stop it")
        server.run_app(app, host, port, path)

    @staticmethod
    async def get_event_type(event: dict):
        """ What the bullshit I made...
        :param event: -> dict
        :return: object
        """
        event_type = Method(event["method"])
        ev = None
        if event_type is Method.PING:
            ev = objects.Ping(**event)

        if event_type is Method.BIND_CHAT:
            ev = objects.BindChat(**event)

        if event_type is Method.BAN_EXPIRED:
            ev = objects.BanExpired(**event)

        if event_type is Method.ADD_USER:
            ev = objects.AddUser(**event)

        if event_type is Method.IGNORE_MESSAGES:
            ev = objects.IgnoreMessages(**event)

        if event_type is Method.SUBSCRIBE_SIGNALS:
            ev = objects.SubscribeSignals(**event)

        if event_type is Method.DELETE_MESSAGES:
            ev = objects.DeleteMessages(**event)

        if event_type is Method.DELETE_MESSAGES_FROM_USER:
            ev = objects.DeleteMessagesFromUser(**event)

        if event_type is Method.PRINT_BOOKMARK:
            ev = objects.PrintBookmark(**event)

        if event_type is Method.FORBIDDEN_LINKS:
            ev = objects.ForbiddenLinks(**event)

        if event_type is Method.SEND_SIGNAL:
            ev = objects.SendSignal(**event)

        if event_type is Method.SEND_MY_SIGNAL:
            ev = objects.SendSignal(**event)

        if event_type is Method.HIRE_API:
            ev = objects.HireApi(**event)

        if event_type is Method.BAN_GET_REASON:
            ev = objects.BanGetReason(**event)

        if event_type is Method.TO_GROUP:
            ev = objects.ToGroup(**event)

        return ev



__version__ = "1.0"

__author__ = "Daniil Tikhonov"

DEFAULT_BOT_FOLDER = "irispy_bot"

errors = {
    1: {"response": "error", "error_code": 1},
    2: {"response": "error", "error_code": 2},
    3: {"response": "error", "error_code": 3},
    4: {"response": "error", "error_code": 4},
    5: {"response": "error", "error_code": 10}
}

prefixes = [".д", ".с", "!дежурный", "!сигнал"]


import typing

from inspect import iscoroutinefunction
from .methods import Method
from ..dispatcher.handler import Handler, SelfHandler
from ..utils import logger


class Event:

    def __init__(self):
        self.handlers: typing.List[Handler] = []
        self.event_handlers: typing.List[SelfHandler] = []

    def __register_event_handler(self, event_type: Method, coro: typing.Callable):
        if not iscoroutinefunction(coro):
            raise Exception("Функция обработчик должна быть асинхронной!")
        handler = Handler(coro, event_type)
        self.handlers.append(handler)
        logger.debug(f"Registered new handler {coro.__name__}")

    def __register_self_handler(
            self,
            event_type: Method,
            coro: typing.Callable,
            commands: typing.List[str],
            lower: bool
    ):
        if not iscoroutinefunction(coro):
            raise Exception("Функция обработчик должна быть асинхронной!")
        handler = SelfHandler(coro, event_type, commands, lower)
        self.event_handlers.append(handler)
        logger.debug(f"Registered new self handler {coro.__name__}")

    def ping(self):
        def decorator(func):
            self.__register_event_handler(Method.PING, func)

        return decorator

    def subscribeSignals(self):
        def decorator(func):
            self.__register_event_handler(Method.SUBSCRIBE_SIGNALS, func)

        return decorator

    def banExpired(self):
        def decorator(func):
            self.__register_event_handler(Method.BAN_EXPIRED, func)

        return decorator

    def addUser(self):
        def decorator(func):
            self.__register_event_handler(Method.ADD_USER, func)

        return decorator

    def deleteMessages(self):
        def decorator(func):
            self.__register_event_handler(Method.DELETE_MESSAGES, func)

        return decorator

    def hireApi(self):
        def decorator(func):
            self.__register_event_handler(Method.HIRE_API, func)

        return decorator

    def deleteMessagesFromUser(self):
        def decorator(func):
            self.__register_event_handler(Method.DELETE_MESSAGES_FROM_USER, func)

        return decorator

    def printBookmark(self):
        def decorator(func):
            self.__register_event_handler(Method.PRINT_BOOKMARK, func)

        return decorator

    def forbiddenLinks(self):
        def decorator(func):
            self.__register_event_handler(Method.FORBIDDEN_LINKS, func)

        return decorator

    def toGroup(self):
        def decorator(func):
            self.__register_event_handler(Method.TO_GROUP, func)

        return decorator

    def banGetReason(self):
        def decorator(func):
            self.__register_event_handler(Method.BAN_GET_REASON, func)

        return decorator

    def bindChat(self):
        def decorator(func):
            self.__register_event_handler(Method.BIND_CHAT, func)

        return decorator

    def ignoreMessages(self):
        def decorator(func):
            self.__register_event_handler(Method.IGNORE_MESSAGES, func)

        return decorator

    def sendMySignal(self, text: typing.Union[str, typing.List[str]], lower: bool = False):
        def decorator(func):
            commands = text if isinstance(text, list) else [text]
            self.__register_self_handler(
                event_type=Method.SEND_MY_SIGNAL,
                coro=func,
                commands=commands,
                lower=lower
            )

        return decorator

    def sendSignal(self, text: typing.Union[str, typing.List[str]], lower: bool = False):
        def decorator(func):
            commands = text if isinstance(text, list) else [text]
            self.__register_self_handler(
                event_type=Method.SEND_SIGNAL,
                coro=func,
                commands=commands,
                lower=lower
            )

        return decorator


import os
import time
import sys


class Logger:
    def __getattr__(self, item):
        if item in ["remove", "add", "level"]:
            return lambda *args, **kwargs: None
        return Logger()

    def __call__(self, message: str, *args, **kwargs):
        t = time.strftime("%m-%d %H:%M:%S", time.localtime())
        sys.stdout.write(
            "\n[IrisPY] " + message.format(*args, **kwargs) + " [TIME {}]".format(t)
        )


def chunks(x, y):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(x), y):
        yield x[i:i + y]


def folder_checkup(path, create: bool = True):
    path = os.path.abspath(path)
    if not os.path.exists(path) and create:
        os.mkdir(path)
    return path


def sub_string(text: str) -> str:
    return " ".join(text.split()[1:])



from enum import Enum


class Method(str, Enum):
    PING = "ping"
    IGNORE_MESSAGES = "ignoreMessages"
    BAN_EXPIRED = "banExpired"
    ADD_USER = "addUser"
    SUBSCRIBE_SIGNALS = "subscribeSignals"
    DELETE_MESSAGES = "deleteMessages"
    DELETE_MESSAGES_FROM_USER = "deleteMessagesFromUser"
    PRINT_BOOKMARK = "printBookmark"
    FORBIDDEN_LINKS = "forbiddenLinks"
    SEND_SIGNAL = "sendSignal"
    SEND_MY_SIGNAL = "sendMySignal"
    HIRE_API = "hireApi"
    BAN_GET_REASON = "banGetReason"
    TO_GROUP = "toGroup"
    BIND_CHAT = "bindChat"


from pydantic import BaseModel


class BaseModel(BaseModel):

    class Config:
        allow_mutation = False
        orm_mode = True

    def __str__(self):
        return str(self.dict())

    def __repr__(self):
        return "%s(%r)" % (self.__class__, self.dict())


from ..utils import logger


class LoggerLevel:
    def __init__(self, level):
        self.level = level

    def __call__(self, record):
        level_no = logger.level(self.level).no
        return record["level"].no >= level_no



from .base import BaseModel
from .methods import Method
from typing import Any, List


class BaseObject(BaseModel):
    user_id: int
    secret: str


class Message(BaseModel):
    conversation_message_id: int
    from_id: int
    date: int
    text: str


class BanExpiredObject(BaseModel):
    user_id: int
    chat: str
    comment: str


class AddUserObject(BaseModel):
    user_id: int
    chat: str


class SubscribeSignalsObject(BaseModel):
    chat: str
    conversation_message_id: int
    text: str
    from_id: int


class DeleteMessagesObject(BaseModel):
    chat: str
    local_ids: List[int]
    is_spam: int


class DeleteMessagesFromUserObject(BaseModel):
    chat: str
    member_ids: List[int]
    user_id: int
    is_spam: int


class IgnoreMessagesObject(BaseModel):
    chat: str
    local_ids: List[int]


class PrintBookmarkObject(BaseModel):
    chat: str
    conversation_message_id: int
    description: int


class ForbiddenLinksObject(BaseModel):
    chat: str
    local_ids: List[int]


class SendSignalObject(BaseModel):
    chat: str
    from_id: int
    value: str
    conversation_message_id: int


class SendMySignalObject(BaseModel):
    chat: str
    from_id: int
    value: str
    conversation_message_id: int


class HireApiObject(BaseModel):
    chat: str
    price: int


class ToGroupObject(BaseModel):
    chat: str
    group_id: int
    local_id: int


class BanGetReasonObject(BaseModel):
    chat: str
    local_id: int


class BindChatObject(BaseModel):
    chat: str


class Ping(BaseObject):
    method: str
    object: Any


class BindChat(BaseObject):
    method: str
    message: Message
    object: BindChatObject


class BanExpired(BaseObject):
    method: str
    object: BanExpiredObject


class AddUser(BaseObject):
    method: str
    object: AddUserObject


class SubscribeSignals(BaseObject):
    method: str
    object: SubscribeSignalsObject


class DeleteMessages(BaseObject):
    method: str
    object: DeleteMessagesObject


class DeleteMessagesFromUser(BaseObject):
    method: str
    message: Message
    object: DeleteMessagesFromUserObject


class IgnoreMessages(BaseObject):
    method: str
    object: IgnoreMessagesObject


class PrintBookmark(BaseObject):
    method: str
    object: PrintBookmarkObject


class ForbiddenLinks(BaseObject):
    method: str
    object: ForbiddenLinksObject


class SendSignal(BaseObject):
    method: str
    message: Message
    object: SendSignalObject


class SendMySignal(BaseObject):
    method: str
    message: Message
    object: SendSignalObject


class HireApi(BaseObject):
    method: str
    object: HireApiObject


class BanGetReason(BaseObject):
    method: str
    object: BanGetReasonObject


class ToGroup(BaseObject):
    method: str
    object: ToGroupObject


BindChat.update_forward_refs()
SendSignal.update_forward_refs()
BanExpired.update_forward_refs()
AddUser.update_forward_refs()
SubscribeSignals.update_forward_refs()
DeleteMessages.update_forward_refs()
DeleteMessagesFromUserObject.update_forward_refs()
PrintBookmark.update_forward_refs()
ForbiddenLinks.update_forward_refs()
SendSignal.update_forward_refs()
SendMySignal.update_forward_refs()
HireApi.update_forward_refs()
BanGetReason.update_forward_refs()
ToGroup.update_forward_refs()






from setuptools import setup

try:
    with open("README.md", "r", encoding="utf-8") as f:
        long_description = f.read()
except (Exception, FileNotFoundError):
    with open("README.md", "r") as f:
        long_description = f.read()

setup(
    name='irispy',
    version='1.1.1',
    packages=['irispy', 'irispy.types', 'irispy.dispatcher', 'irispy.utils'],
    url='https://github.com/zpodushkin/irispy',
    download_url='https://github.com/zpodushkin/irispy/archive/1.1.tar.gz',
    license='GPL-3.0',
    author='zpodushkin',
    description='Async and fast api wrapper for IrisCallback API',
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Operating System :: OS Independent",
    ],
    install_requires=['aiohttp', 'pydantic', 'loguru', 'vbml']
)
